from __future__ import annotations

from django.core.mail import EmailMessage
from django.core.mail import EmailMultiAlternatives

from email_relay.email import RelayEmailData


def test_from_email_message():
    email_message = EmailMessage(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )

    relay_email_data = RelayEmailData.from_email_message(email_message)

    assert relay_email_data.subject == email_message.subject
    assert relay_email_data.body == email_message.body
    assert relay_email_data.from_email == email_message.from_email
    assert relay_email_data.to == email_message.to
    assert relay_email_data.cc == email_message.cc
    assert relay_email_data.bcc == email_message.bcc
    assert relay_email_data.reply_to == email_message.reply_to
    assert relay_email_data.extra_headers == email_message.extra_headers
    assert relay_email_data.alternatives == []
    assert relay_email_data.attachments == []


def test_from_email_message_multi_alternatives():
    email_multi_alternatives = EmailMultiAlternatives(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )
    email_multi_alternatives.attach_alternative(
        "<p>Here is the message.</p>", "text/html"
    )

    relay_email_data = RelayEmailData.from_email_message(email_multi_alternatives)

    assert relay_email_data.subject == email_multi_alternatives.subject
    assert relay_email_data.body == email_multi_alternatives.body
    assert relay_email_data.from_email == email_multi_alternatives.from_email
    assert relay_email_data.to == email_multi_alternatives.to
    assert relay_email_data.cc == email_multi_alternatives.cc
    assert relay_email_data.bcc == email_multi_alternatives.bcc
    assert relay_email_data.reply_to == email_multi_alternatives.reply_to
    assert relay_email_data.extra_headers == email_multi_alternatives.extra_headers
    assert relay_email_data.alternatives == email_multi_alternatives.alternatives
    assert relay_email_data.attachments == []


def test_to_dict():
    email_message = EmailMessage(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )

    email_dict = RelayEmailData.from_email_message(email_message).to_dict()

    assert email_dict == {
        "subject": email_message.subject,
        "body": email_message.body,
        "from_email": email_message.from_email,
        "to": email_message.to,
        "cc": email_message.cc,
        "bcc": email_message.bcc,
        "reply_to": email_message.reply_to,
        "extra_headers": email_message.extra_headers,
        "alternatives": [],
        "attachments": [],
    }


def test_to_dict_multi_alternatives():
    email_multi_alternatives = EmailMultiAlternatives(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )
    email_multi_alternatives.attach_alternative(
        "<p>Here is the message.</p>", "text/html"
    )

    email_dict = RelayEmailData.from_email_message(email_multi_alternatives).to_dict()

    assert email_dict == {
        "subject": email_multi_alternatives.subject,
        "body": email_multi_alternatives.body,
        "from_email": email_multi_alternatives.from_email,
        "to": email_multi_alternatives.to,
        "cc": email_multi_alternatives.cc,
        "bcc": email_multi_alternatives.bcc,
        "reply_to": email_multi_alternatives.reply_to,
        "extra_headers": email_multi_alternatives.extra_headers,
        "alternatives": email_multi_alternatives.alternatives,
        "attachments": [],
    }


def test_to_dict_with_attachment():
    email = EmailMessage(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )
    attachment_content = b"Hello World!"
    email.attach(
        filename="test.txt",
        content=attachment_content,
        mimetype="text/plain",
    )

    email_dict = RelayEmailData.from_email_message(email).to_dict()

    assert email_dict == {
        "subject": email.subject,
        "body": email.body,
        "from_email": email.from_email,
        "to": email.to,
        "cc": email.cc,
        "bcc": email.bcc,
        "reply_to": email.reply_to,
        "extra_headers": email.extra_headers,
        "alternatives": [],
        "attachments": [
            {
                "filename": "test.txt",
                "content": attachment_content.decode(),
                "mimetype": "text/plain",
            }
        ],
    }



from __future__ import annotations

import base64
import binascii
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from email.mime.base import MIMEBase

from django.core.mail import EmailMessage
from django.core.mail import EmailMultiAlternatives


@dataclass(frozen=True)
class RelayEmailData:
    subject: str = ""
    body: str = ""
    from_email: str = ""
    to: list[str] = field(default_factory=list)
    cc: list[str] = field(default_factory=list)
    bcc: list[str] = field(default_factory=list)
    reply_to: list[str] = field(default_factory=list)
    extra_headers: dict[str, str] = field(default_factory=dict)
    alternatives: list[tuple[str, str]] = field(default_factory=list)
    attachments: list[dict[str, str]] = field(default_factory=list)

    def to_dict(self) -> dict[str, str]:
        return asdict(self)

    def to_email_message(self) -> EmailMultiAlternatives:
        email = EmailMultiAlternatives(
            subject=self.subject,
            body=self.body,
            from_email=self.from_email,
            to=self.to,
            cc=self.cc,
            bcc=self.bcc,
            reply_to=self.reply_to,
            headers=self.extra_headers,
        )

        for alternative in self.alternatives:
            email.attach_alternative(alternative[0], alternative[1])

        for attachment in self.attachments:
            content = attachment.get("content", "")
            try:
                # Attempt to decode the base64 string into bytes
                decoded_content = base64.b64decode(content)
            except binascii.Error:
                # Fallback to assuming it's plain text, encoded as bytes
                decoded_content = content.encode("utf-8")

            email.attach(
                filename=attachment.get("filename", ""),
                content=decoded_content,
                mimetype=attachment.get("mimetype", ""),
            )

        return email

    @classmethod
    def from_email_message(
        cls, email_message: EmailMessage | EmailMultiAlternatives
    ) -> RelayEmailData:
        attachments = []
        for attachment in email_message.attachments:
            if isinstance(attachment, MIMEBase):
                attachments.append(
                    {
                        "filename": attachment.get_filename(
                            failobj="filename_not_found"
                        ),
                        "content": base64.b64encode(
                            attachment.get_payload(decode=True)
                        ).decode(),
                        "mimetype": attachment.get_content_type(),
                    }
                )
            else:
                if isinstance(attachment[1], bytes):
                    content = base64.b64encode(attachment[1]).decode("utf-8")
                else:
                    content = attachment[1]

                attachments.append(
                    {
                        "filename": attachment[0],
                        "content": content,
                        "mimetype": attachment[2],
                    }
                )

        return cls(
            subject=str(email_message.subject),
            body=str(email_message.body),
            from_email=email_message.from_email,
            to=email_message.to,
            cc=email_message.cc,
            bcc=email_message.bcc,
            reply_to=email_message.reply_to,
            extra_headers=email_message.extra_headers,
            alternatives=getattr(email_message, "alternatives", []),
            attachments=attachments,
        )



# Generated by Django 4.2.5 on 2023-10-30 13:04

from django.db import migrations


def migrate_message_data_to_new_schema(apps, schema_editor):
    """Migrate the JSON data from the old schema to the new schema.

    Changes:
        - "message" is now "body"
        - "recipient_list" is now "to"
        - "html_message" is now a part of "alternatives", a list of tuples of (content, mimetype)

    Adds:
        - "cc"
        - "bcc"
        - "reply_to"
        - "extra_headers"
        - "alternatives"
    """
    Message = apps.get_model("email_relay", "Message")

    for message in Message.objects.all():
        data = message.data
        if not data:
            continue
        data["body"] = data.pop("message", "")
        data["to"] = data.pop("recipient_list", [])
        data["cc"] = []
        data["bcc"] = []
        data["reply_to"] = []
        data["extra_headers"] = {}
        data["alternatives"] = []
        html_message = data.pop("html_message", None)
        if html_message:
            data["alternatives"].append((html_message, "text/html"))
        message.save()


class Migration(migrations.Migration):
    dependencies = [
        ("email_relay", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(
            migrate_message_data_to_new_schema, migrations.RunPython.noop
        ),
    ]



from __future__ import annotations

import pytest
from django.conf import settings
from django.core.mail import EmailMessage
from django.core.mail import EmailMultiAlternatives
from django.core.mail import mail_admins
from django.core.mail import mail_managers
from django.core.mail import send_mail
from django.core.mail import send_mass_mail
from django.test import override_settings

from email_relay.conf import EMAIL_RELAY_DATABASE_ALIAS
from email_relay.models import Message

pytestmark = pytest.mark.django_db(databases=["default", EMAIL_RELAY_DATABASE_ALIAS])


@pytest.fixture(autouse=True, scope="module")
def relay_email_backend():
    with override_settings(
        EMAIL_BACKEND="email_relay.backend.RelayDatabaseEmailBackend",
    ):
        yield


def test_send_mail(mailoutbox):
    send_mail(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        html_message="<p>Here is the message.</p>",
    )

    assert Message.objects.count() == 1
    assert len(mailoutbox) == 0

    email = Message.objects.get().email
    assert email.subject == "Subject here"
    assert email.body == "Here is the message."
    assert email.from_email == "from@example.com"
    assert email.to == ["to@example.com"]
    assert email.alternatives == [("<p>Here is the message.</p>", "text/html")]


def test_send_mass_mail(mailoutbox):
    send_mass_mail(
        (
            (
                "Subject here",
                "Here is the message.",
                "from@example.com",
                ["to@example.com"],
            ),
            (
                "Another subject",
                "Here is another message.",
                "from@example.com",
                ["to@example.com"],
            ),
        ),
    )

    assert Message.objects.count() == 2
    assert len(mailoutbox) == 0


@override_settings(ADMINS=[("Admin", "admin@example.com")])
def test_mail_admins(mailoutbox):
    mail_admins(
        "Subject here",
        "Here is the message.",
    )

    assert Message.objects.count() == 1
    assert len(mailoutbox) == 0

    email = Message.objects.get().email
    assert email.subject == f"{settings.EMAIL_SUBJECT_PREFIX}Subject here"
    assert email.body == "Here is the message."
    assert email.from_email == settings.SERVER_EMAIL
    assert email.to == ["admin@example.com"]


@override_settings(MANAGERS=[("Manager", "manager@example.com")])
def test_mail_managers(mailoutbox):
    mail_managers(
        "Subject here",
        "Here is the message.",
    )

    assert Message.objects.count() == 1
    assert len(mailoutbox) == 0

    email = Message.objects.get().email
    assert email.subject == f"{settings.EMAIL_SUBJECT_PREFIX}Subject here"
    assert email.body == "Here is the message."
    assert email.from_email == settings.SERVER_EMAIL
    assert email.to == ["manager@example.com"]


def test_emailmessage(mailoutbox):
    email_message = EmailMessage(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )

    email_message.send()

    assert Message.objects.count() == 1
    assert len(mailoutbox) == 0

    email = Message.objects.get().email
    assert email.subject == "Subject here"
    assert email.body == "Here is the message."
    assert email.from_email == "from@example.com"
    assert email.to == ["to@example.com"]
    assert email.cc == ["cc@example.com"]
    assert email.bcc == ["bcc@example.com"]
    assert email.reply_to == ["reply_to@example.com"]
    assert email.extra_headers == {"Test-Header": "Test Value"}


def test_emailmessage_attach(mailoutbox):
    email_message = EmailMessage(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )
    email_message.attach("attachment.txt", "Here is the attachment.")

    email_message.send()

    assert Message.objects.count() == 1
    assert len(mailoutbox) == 0

    email = Message.objects.get().email
    assert email.subject == "Subject here"
    assert email.body == "Here is the message."
    assert email.from_email == "from@example.com"
    assert email.to == ["to@example.com"]
    assert email.cc == ["cc@example.com"]
    assert email.bcc == ["bcc@example.com"]
    assert email.reply_to == ["reply_to@example.com"]
    assert email.extra_headers == {"Test-Header": "Test Value"}
    assert email.attachments == [
        ("attachment.txt", "Here is the attachment.", "text/plain"),
    ]


def test_emailmessage_attach_file(tmp_path, mailoutbox):
    email_message = EmailMessage(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )
    file_path = tmp_path / "attachment.txt"
    file_path.write_text("Here is the attachment.")
    email_message.attach_file(str(file_path))

    email_message.send()

    assert Message.objects.count() == 1
    assert len(mailoutbox) == 0

    email = Message.objects.get().email
    assert email.subject == "Subject here"
    assert email.body == "Here is the message."
    assert email.from_email == "from@example.com"
    assert email.to == ["to@example.com"]
    assert email.cc == ["cc@example.com"]
    assert email.bcc == ["bcc@example.com"]
    assert email.reply_to == ["reply_to@example.com"]
    assert email.extra_headers == {"Test-Header": "Test Value"}
    assert email.attachments == [
        ("attachment.txt", "Here is the attachment.", "text/plain"),
    ]


def test_emailmessagealternatives(mailoutbox):
    email_multi_alternatives = EmailMultiAlternatives(
        "Subject here",
        "Here is the message.",
        "from@example.com",
        ["to@example.com"],
        cc=["cc@example.com"],
        bcc=["bcc@example.com"],
        reply_to=["reply_to@example.com"],
        headers={"Test-Header": "Test Value"},
    )
    email_multi_alternatives.attach_alternative(
        "<p>Here is the message.</p>", "text/html"
    )

    email_multi_alternatives.send()

    assert Message.objects.count() == 1
    assert len(mailoutbox) == 0

    email = Message.objects.get().email
    assert email.subject == "Subject here"
    assert email.body == "Here is the message."
    assert email.from_email == "from@example.com"
    assert email.to == ["to@example.com"]
    assert email.cc == ["cc@example.com"]
    assert email.bcc == ["bcc@example.com"]
    assert email.reply_to == ["reply_to@example.com"]
    assert email.extra_headers == {"Test-Header": "Test Value"}
    assert email.alternatives == [
        ("<p>Here is the message.</p>", "text/html"),
    ]



from __future__ import annotations

import importlib

import pytest
from django.apps import apps
from django.db import connections
from model_bakery import baker

from email_relay.conf import EMAIL_RELAY_DATABASE_ALIAS
from email_relay.models import Message


@pytest.fixture
def migrate_message_data_to_new_schema():
    return importlib.import_module(
        "email_relay.migrations.0002_auto_20231030_1304"
    ).migrate_message_data_to_new_schema


@pytest.mark.django_db(databases=["default", EMAIL_RELAY_DATABASE_ALIAS])
def test_migrate_message_data_to_new_schema(migrate_message_data_to_new_schema):
    class MockSchemaEditor:
        connection = connections[EMAIL_RELAY_DATABASE_ALIAS]

    baker.make(
        "email_relay.Message",
        data={
            "message": "Here is the message.",
            "recipient_list": ["to@example.com"],
            "html_message": "<p>HTML</p>",
        },
        _quantity=3,
    )

    for message in Message.objects.all():
        assert message.data["message"] == "Here is the message."
        assert message.data["recipient_list"] == ["to@example.com"]
        assert message.data["html_message"] == "<p>HTML</p>"
        assert not message.data.get("to")
        assert not message.data.get("cc")
        assert not message.data.get("bcc")
        assert not message.data.get("reply_to")
        assert not message.data.get("extra_headers")
        assert not message.data.get("alternatives")

    migrate_message_data_to_new_schema(apps, MockSchemaEditor())

    assert Message.objects.count() == 3

    for message in Message.objects.all():
        assert not message.data.get("message")
        assert not message.data.get("recipient_list")
        assert not message.data.get("html_message")
        assert message.data["body"] == "Here is the message."
        assert message.data["to"] == ["to@example.com"]
        assert message.data["cc"] == []
        assert message.data["bcc"] == []
        assert message.data["reply_to"] == []
        assert message.data["extra_headers"] == {}
        assert message.data["alternatives"] == [
            ["<p>HTML</p>", "text/html"],
        ]



from __future__ import annotations

import datetime
import logging
from itertools import chain

from django.core.mail import EmailMessage
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.utils import timezone

from email_relay.conf import app_settings
from email_relay.email import RelayEmailData

logger = logging.getLogger(__name__)


class Priority(models.IntegerChoices):
    LOW = 1, "Low"
    MEDIUM = 2, "Medium"
    HIGH = 3, "High"


class Status(models.IntegerChoices):
    QUEUED = 1, "Queued"
    DEFERRED = 2, "Deferred"
    FAILED = 3, "Failed"
    SENT = 4, "Sent"


class MessageManager(models.Manager["Message"]):
    def get_message_batch(self) -> list[Message]:
        message_batch = list(
            chain(
                self.queued().prioritized(),  # type: ignore[attr-defined]
                self.deferred().prioritized(),  # type: ignore[attr-defined]
            )
        )
        logger.debug(f"found {len(message_batch)} messages to send")
        if app_settings.EMAIL_MAX_BATCH is not None:
            msg = f"max batch size is {app_settings.EMAIL_MAX_BATCH}"
            if len(message_batch) > app_settings.EMAIL_MAX_BATCH:
                msg += ", truncating"
            logger.debug(msg)
            message_batch = message_batch[: app_settings.EMAIL_MAX_BATCH]
        return message_batch

    def get_message_for_sending(self, id: int) -> Message:
        return self.filter(id=id).select_for_update(skip_locked=True).get()

    def messages_available_to_send(self) -> bool:
        return self.queued().exists() or self.deferred().exists()  # type: ignore[attr-defined]

    def delete_all_sent_messages(self) -> int:
        return self.sent().delete()[0]  # type: ignore[attr-defined]

    def delete_messages_sent_before(self, dt: datetime.datetime) -> int:
        return self.sent_before(dt).delete()[0]  # type: ignore[attr-defined]


class MessageQuerySet(models.QuerySet["Message"]):
    def prioritized(self):
        return self.order_by("-priority", "created_at")

    def high_priority(self):
        return self.filter(priority=Priority.HIGH)

    def medium_priority(self):
        return self.filter(priority=Priority.MEDIUM)

    def low_priority(self):
        return self.filter(priority=Priority.LOW)

    def queued(self):
        return self.filter(status=Status.QUEUED)

    def deferred(self):
        return self.filter(status=Status.DEFERRED)

    def failed(self):
        return self.filter(status=Status.FAILED)

    def sent(self):
        return self.filter(status=Status.SENT)

    def sent_before(self, dt: datetime.datetime):
        return self.sent().filter(sent_at__lte=dt)


# This is a workaround to make `mypy` happy
_MessageManager = MessageManager.from_queryset(MessageQuerySet)


class Message(models.Model):
    id: int
    data = models.JSONField()
    priority = models.PositiveSmallIntegerField(
        choices=Priority.choices, default=Priority.LOW
    )
    status = models.PositiveSmallIntegerField(
        choices=Status.choices, default=Status.QUEUED
    )
    retry_count = models.PositiveSmallIntegerField(default=0)
    log = models.TextField(
        blank=True, help_text="Most recent log message from the email backend, if any."
    )

    created_at = models.DateTimeField(auto_now_add=True, editable=False)
    updated_at = models.DateTimeField(auto_now=True, editable=False)
    sent_at = models.DateTimeField(null=True, blank=True)

    objects = _MessageManager()

    class Meta:
        ordering = ["created_at"]

    def __str__(self):
        try:
            return f'{self.created_at} "{self.data["subject"]}" to {", ".join(self.data["to"])}'
        except Exception:
            return f"{self.created_at} <invalid message>"

    def save(self, *args, **kwargs):
        # Overriding the save method in order to make sure that
        # modified field is updated even if it is not given as
        # a parameter to the update field argument.
        update_fields = kwargs.get("update_fields", None)
        if update_fields:
            kwargs["update_fields"] = set(update_fields).union({"updated_at"})

        super().save(*args, **kwargs)

    def mark_sent(self):
        self.status = Status.SENT
        self.sent_at = timezone.now()
        self.save()

    def defer(self, log: str = ""):
        self.status = Status.DEFERRED
        self.log = log
        self.retry_count += 1
        self.save()

    def fail(self, log: str = ""):
        self.status = Status.FAILED
        self.log = log
        self.save()

    @property
    def email(self) -> EmailMultiAlternatives | None:
        data = self.data
        if not data:
            return None

        return RelayEmailData(**data).to_email_message()

    @email.setter
    def email(self, email_message: EmailMessage | EmailMultiAlternatives) -> None:
        self.data = RelayEmailData.from_email_message(email_message).to_dict()



from __future__ import annotations

import pytest
from django.conf import settings
from django.test import override_settings

from email_relay.conf import app_settings


@pytest.mark.parametrize(
    "setting,default_setting",
    [
        ("DATABASE_ALIAS", "email_relay_db"),
        ("EMAIL_MAX_BATCH", None),
        ("EMAIL_MAX_DEFERRED", None),
        ("EMAIL_MAX_RETRIES", None),
        ("EMPTY_QUEUE_SLEEP", 30),
        ("EMAIL_THROTTLE", 0),
        ("MESSAGES_BATCH_SIZE", None),
        ("MESSAGES_RETENTION_SECONDS", None),
        ("RELAY_HEALTHCHECK_METHOD", "GET"),
        ("RELAY_HEALTHCHECK_STATUS_CODE", 200),
        ("RELAY_HEALTHCHECK_TIMEOUT", 5.0),
        ("RELAY_HEALTHCHECK_URL", None),
    ],
)
def test_default_settings(setting, default_setting):
    user_settings = getattr(settings, "DJANGO_EMAIL_RELAY", {})

    assert user_settings == {}
    assert getattr(app_settings, setting) == default_setting


@pytest.mark.parametrize(
    "setting,user_setting",
    [
        ("DATABASE_ALIAS", "custom_db_name"),
        ("EMAIL_MAX_BATCH", 10),
        ("EMAIL_MAX_DEFERRED", 10),
        ("EMAIL_MAX_RETRIES", 10),
        ("EMPTY_QUEUE_SLEEP", 1),
        ("EMAIL_THROTTLE", 1),
        ("MESSAGES_BATCH_SIZE", 10),
        ("MESSAGES_RETENTION_SECONDS", 10),
        ("RELAY_HEALTHCHECK_METHOD", "POST"),
        ("RELAY_HEALTHCHECK_STATUS_CODE", 201),
        ("RELAY_HEALTHCHECK_TIMEOUT", 10.0),
        ("RELAY_HEALTHCHECK_URL", "http://example.com/healthcheck"),
    ],
)
def test_custom_settings(setting, user_setting):
    with override_settings(
        DJANGO_EMAIL_RELAY={
            setting: user_setting,
        },
    ):
        assert getattr(app_settings, setting) == user_setting



from __future__ import annotations

import base64
import datetime
from email.mime.base import MIMEBase

import pytest
from django.core.mail import EmailMessage
from django.core.mail import EmailMultiAlternatives
from django.test import override_settings
from django.utils import timezone
from model_bakery import baker

from email_relay.models import Message
from email_relay.models import Priority
from email_relay.models import Status


@pytest.mark.django_db(databases=["default", "email_relay_db"])
def test_message():
    baker.make("email_relay.Message")
    assert Message.objects.all().count() == 1


@pytest.mark.django_db(databases=["default", "email_relay_db"])
class TestMessageManager:
    def test_get_message_batch(self):
        baker.make("email_relay.Message", status=Status.QUEUED, _quantity=5)
        baker.make("email_relay.Message", status=Status.DEFERRED, _quantity=5)

        message_batch = Message.objects.get_message_batch()

        assert len(message_batch) == 10

    @override_settings(
        DJANGO_EMAIL_RELAY={
            "EMAIL_MAX_BATCH": 1,
        }
    )
    def test_get_message_batch_with_max_batch_size(self):
        baker.make("email_relay.Message", status=Status.QUEUED, _quantity=5)
        baker.make("email_relay.Message", status=Status.DEFERRED, _quantity=5)

        message_batch = Message.objects.get_message_batch()

        assert len(message_batch) == 1

    def test_get_message_for_sending(self):
        message = baker.make("email_relay.Message", status=Status.QUEUED)

        message_for_sending = Message.objects.get_message_for_sending(message.id)

        assert message_for_sending == message

    @pytest.mark.parametrize(
        "status, expected",
        [
            (Status.QUEUED, True),
            (Status.DEFERRED, True),
            (Status.FAILED, False),
            (Status.SENT, False),
        ],
    )
    def test_messages_available_to_send(self, status, expected):
        baker.make("email_relay.Message", status=status)

        assert Message.objects.messages_available_to_send() == expected

    def test_messages_available_to_send_with_no_messages(self):
        assert not Message.objects.messages_available_to_send()

    def test_delete_all_sent_messages(self):
        baker.make("email_relay.Message", status=Status.SENT, _quantity=5)

        deleted_messages = Message.objects.delete_all_sent_messages()

        assert deleted_messages == 5
        assert Message.objects.count() == 0

    def test_delete_messages_sent_before(self):
        one_week = baker.make(
            "email_relay.Message",
            status=Status.SENT,
            sent_at=timezone.now() - datetime.timedelta(days=7),
        )
        now = baker.make(
            "email_relay.Message",
            status=Status.SENT,
            sent_at=timezone.now(),
        )
        not_sent = baker.make(
            "email_relay.Message",
            status=Status.QUEUED,
            sent_at=None,
        )

        deleted_messages = Message.objects.delete_messages_sent_before(
            timezone.now() - datetime.timedelta(days=1)
        )

        assert deleted_messages == 1
        assert Message.objects.count() == 2

        messages = Message.objects.all()
        assert one_week not in messages
        assert now in messages
        assert not_sent in messages


@pytest.mark.django_db(databases=["default", "email_relay_db"])
class TestMessageQuerySet:
    @pytest.fixture
    def messages_with_priority(self):
        low = baker.make("email_relay.Message", priority=Priority.LOW)
        medium = baker.make("email_relay.Message", priority=Priority.MEDIUM)
        high = baker.make("email_relay.Message", priority=Priority.HIGH)
        return {
            "low": low,
            "medium": medium,
            "high": high,
        }

    @pytest.fixture
    def messages_with_status(self):
        queued = baker.make("email_relay.Message", status=Status.QUEUED)
        deferred = baker.make("email_relay.Message", status=Status.DEFERRED)
        failed = baker.make("email_relay.Message", status=Status.FAILED)
        sent = baker.make("email_relay.Message", status=Status.SENT)
        return {
            "queued": queued,
            "deferred": deferred,
            "failed": failed,
            "sent": sent,
        }

    def test_prioritized(self, messages_with_priority):
        queryset = Message.objects.prioritized()

        assert queryset.count() == 3
        assert queryset[0] == messages_with_priority["high"]
        assert queryset[1] == messages_with_priority["medium"]
        assert queryset[2] == messages_with_priority["low"]

    def test_high_priority(self, messages_with_priority):
        queryset = Message.objects.high_priority()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_priority["high"]

    def test_medium_priority(self, messages_with_priority):
        queryset = Message.objects.medium_priority()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_priority["medium"]

    def test_low_priority(self, messages_with_priority):
        queryset = Message.objects.low_priority()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_priority["low"]

    def test_queued(self, messages_with_status):
        queryset = Message.objects.queued()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["queued"]

    def test_deferred(self, messages_with_status):
        queryset = Message.objects.deferred()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["deferred"]

    def test_failed(self, messages_with_status):
        queryset = Message.objects.failed()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["failed"]

    def test_sent(self, messages_with_status):
        queryset = Message.objects.sent()

        assert queryset.count() == 1
        assert queryset[0] == messages_with_status["sent"]

    def test_sent_before(self):
        one_week = baker.make(
            "email_relay.Message",
            status=Status.SENT,
            sent_at=timezone.now() - datetime.timedelta(days=7),
        )
        now = baker.make(
            "email_relay.Message",
            status=Status.SENT,
            sent_at=timezone.now(),
        )
        not_sent = baker.make(
            "email_relay.Message",
            status=Status.QUEUED,
            sent_at=None,
        )

        queryset = Message.objects.sent_before(
            timezone.now() - datetime.timedelta(days=1)
        )

        assert queryset.count() == 1
        assert one_week in queryset
        assert now not in queryset
        assert not_sent not in queryset


@pytest.mark.django_db(databases=["default", "email_relay_db"])
class TestMessageModel:
    @pytest.fixture
    def data(self):
        return {
            "subject": "Test",
            "body": "Test",
            "from_email": "from@example.com",
            "to": ["to@example.com"],
        }

    @pytest.fixture
    def email(self):
        return EmailMultiAlternatives(
            subject="Test",
            body="Test",
            from_email="from@example.com",
            to=["to@example.com"],
        )

    @pytest.fixture
    def queued_message(self, data):
        return baker.make("email_relay.Message", data=data, status=Status.QUEUED)

    def test_create(self, data):
        message = baker.make("email_relay.Message", data=data)

        assert message.data == data
        assert message.priority == Priority.LOW
        assert message.status == Status.QUEUED
        assert message.retry_count == 0
        assert message.log == ""
        assert message.sent_at is None

    def test_str(self, data):
        message = baker.make("email_relay.Message", data=data)

        assert data["subject"] in str(message)

    def test_str_invalid_data(self):
        message = baker.make("email_relay.Message", data={})

        assert "invalid message" in str(message)

    def test_update_with_update_fields(self, data):
        message = baker.make("email_relay.Message", data=data)
        updated_at_original = message.updated_at

        message.retry_count = 1
        message.save(update_fields=["retry_count"])

        assert message.updated_at != updated_at_original

    def test_mark_sent(self, queued_message):
        queued_message.mark_sent()

        assert queued_message.status == Status.SENT

    def test_defer(self, queued_message):
        queued_message.defer()

        assert queued_message.status == Status.DEFERRED

    def test_fail(self, queued_message):
        queued_message.fail()

        assert queued_message.status == Status.FAILED

    def test_no_data(self):
        message = baker.make("email_relay.Message", data={})

        assert message.data == {}
        assert message.email is None

    def test_email_property(self, data):
        message = Message.objects.create(data=data)

        email = message.email

        assert isinstance(email, EmailMessage)
        assert email.subject == data["subject"]
        assert email.body == data["body"]
        assert email.from_email == data["from_email"]
        assert email.to == data["to"]

    def test_email_setter(self, data):
        message = Message.objects.create(data=data)
        email = EmailMultiAlternatives(
            subject="Test 2",
            body="Test 2",
            from_email="from2@example.com",
            to=["to2@example.com"],
        )

        message.email = email
        message.save()

        assert message.data["subject"] == email.subject
        assert message.data["body"] == email.body
        assert message.data["from_email"] == email.from_email
        assert message.data["to"] == email.to

    def test_email_with_plain_text_attachment(self, email):
        attachment_content = b"Hello World!"
        email.attach(
            filename="test.txt",
            content=attachment_content,
            mimetype="text/plain",
        )

        message = Message()
        message.email = email
        message.save()

        assert Message.objects.count() == 1

        saved_message = Message.objects.first()
        assert saved_message.data["attachments"][0]["filename"] == "test.txt"
        assert saved_message.data["attachments"][0][
            "content"
        ] == attachment_content.decode("utf-8")
        assert saved_message.data["attachments"][0]["mimetype"] == "text/plain"

        email_from_db = saved_message.email
        assert email_from_db.attachments[0][0] == "test.txt"
        assert email_from_db.attachments[0][1] == attachment_content.decode("utf-8")
        assert email_from_db.attachments[0][2] == "text/plain"

    def test_email_with_binary_attachment(self, email, faker):
        attachment_content = faker.binary(length=10)
        email.attach(
            filename="test.zip",
            content=attachment_content,
            mimetype="application/zip",
        )

        message = Message()
        message.email = email
        message.save()

        assert Message.objects.count() == 1

        saved_message = Message.objects.first()
        assert saved_message.data["attachments"][0]["filename"] == "test.zip"
        assert saved_message.data["attachments"][0]["content"] == base64.b64encode(
            attachment_content
        ).decode("utf-8")
        assert saved_message.data["attachments"][0]["mimetype"] == "application/zip"

        email_from_db = saved_message.email
        assert email_from_db.attachments[0][0] == "test.zip"
        assert email_from_db.attachments[0][1] == attachment_content
        assert email_from_db.attachments[0][2] == "application/zip"

    def test_email_with_mimebase_attachment(self, email):
        attachment_content = b"Hello World!"
        attachment = MIMEBase("application", "octet-stream")
        attachment["Content-Disposition"] = 'attachment; filename="test.txt"'
        attachment.set_payload(attachment_content)
        email.attach(attachment)

        message = Message()
        message.email = email
        message.save()

        assert Message.objects.count() == 1

        saved_message = Message.objects.first()
        assert saved_message.data["attachments"][0]["filename"] == "test.txt"
        assert saved_message.data["attachments"][0]["content"] == base64.b64encode(
            attachment_content
        ).decode("utf-8")
        assert (
            saved_message.data["attachments"][0]["mimetype"]
            == "application/octet-stream"
        )

        email_from_db = saved_message.email
        assert email_from_db.attachments[0][0] == "test.txt"
        assert email_from_db.attachments[0][1] == attachment_content
        assert email_from_db.attachments[0][2] == "application/octet-stream"

    def test_email_send(self, email, mailoutbox):
        message = Message()
        message.email = email
        message.save()

        message.email.send()

        assert len(mailoutbox) == 1

    def test_email_send_with_plain_text_attachment(self, email, mailoutbox):
        email.attach(
            filename="test.txt",
            content=b"Hello World!",
            mimetype="text/plain",
        )
        message = Message()
        message.email = email
        message.save()

        message.email.send()

        assert len(mailoutbox) == 1

    def test_email_send_with_binary_attachment(self, email, faker, mailoutbox):
        email.attach(
            filename="test.zip",
            content=faker.binary(length=10),
            mimetype="application/zip",
        )
        message = Message()
        message.email = email
        message.save()

        message.email.send()

        assert len(mailoutbox) == 1

    def test_email_send_with_mimebase_attachment(self, email, mailoutbox):
        attachment = MIMEBase("application", "octet-stream")
        attachment["Content-Disposition"] = 'attachment; filename="test.txt"'
        attachment.set_payload(b"Hello World!")
        email.attach(attachment)

        message = Message()
        message.email = email
        message.save()

        message.email.send()

        assert len(mailoutbox) == 1



from __future__ import annotations

import datetime
import logging
import time

from django.core.management import BaseCommand
from django.utils import timezone

from email_relay.conf import app_settings
from email_relay.models import Message
from email_relay.relay import send_all

try:
    import requests
except ImportError:  # pragma: no cover
    requests = None  # type: ignore[assignment]

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def handle(self, *args, _loop_count: int | None = None, **options) -> None:
        # _loop_count is used to make testing a bit easier
        # it is not intended to be used in production
        loop_count = 0 if _loop_count is not None else None

        logger.info("starting relay")

        while True:
            if Message.objects.messages_available_to_send():
                send_all()

            self.delete_old_messages()
            self.ping_healthcheck()

            msg = "loop complete"
            if app_settings.EMPTY_QUEUE_SLEEP > 0:
                msg += f", sleeping for {app_settings.EMPTY_QUEUE_SLEEP} seconds before next loop"
            logger.debug(msg)

            if _loop_count is not None and loop_count is not None:
                loop_count += 1
                if loop_count >= _loop_count:
                    break

            time.sleep(app_settings.EMPTY_QUEUE_SLEEP)

    def delete_old_messages(self) -> None:
        if app_settings.MESSAGES_RETENTION_SECONDS is not None:
            logger.debug("deleting old messages")
            if app_settings.MESSAGES_RETENTION_SECONDS == 0:
                deleted_messages = Message.objects.delete_all_sent_messages()
            else:
                deleted_messages = Message.objects.delete_messages_sent_before(
                    timezone.now()
                    - datetime.timedelta(
                        seconds=app_settings.MESSAGES_RETENTION_SECONDS
                    )
                )
            logger.debug(f"deleted {deleted_messages} messages")

    def ping_healthcheck(self) -> None:
        if app_settings.RELAY_HEALTHCHECK_URL is not None:
            if requests is None:
                logger.warning(
                    "Healthcheck URL configured but requests is not installed. "
                    "Please install requests to use the healthcheck feature."
                )
                return

            logger.debug("pinging healthcheck")
            try:
                response = requests.request(
                    method=app_settings.RELAY_HEALTHCHECK_METHOD,
                    url=app_settings.RELAY_HEALTHCHECK_URL,
                    timeout=app_settings.RELAY_HEALTHCHECK_TIMEOUT,
                )
            except requests.exceptions.RequestException as e:
                logger.warning(f"healthcheck failed, got exception: {e}")
                return

            if response.status_code == app_settings.RELAY_HEALTHCHECK_STATUS_CODE:
                logger.debug("healthcheck ping successful")
            else:
                logger.warning(
                    f"healthcheck failed, got {response.status_code}, expected {app_settings.RELAY_HEALTHCHECK_STATUS_CODE}"
                )
